@@ -21,8 +21,6 @@
 // Color space transformation shaders. These all input and output a color
 // value (PL_SHADER_SIG_COLOR).
 
-#include <stdint.h>
-
 #include <libplacebo/colorspace.h>
 #include <libplacebo/shaders.h>
 
@@ -143,7 +141,7 @@ bool pl_shader_detect_peak(struct pl_shader *sh,
 // function is *not* needed when the user only wants to use `pl_shader_color_map`,
 // since that can ingest the peak detection state object directly. It only
 // serves as a utility/debugging function.
-bool pl_get_detected_peak(const struct pl_shader_obj **state,
+bool pl_get_detected_peak(const struct pl_shader_obj *state,
                           float *out_peak, float *out_avg);
 
 // A collection of various tone mapping algorithms supported by libplacebo.
@@ -204,6 +202,8 @@ enum pl_tone_mapping_algorithm {
     // EETF from the ITU-R Report BT.2390, a hermite spline roll-off with
     // linear segment. Not configurable.
     PL_TONE_MAPPING_BT_2390,
+
+    PL_TONE_MAPPING_ALGORITHM_COUNT,
 };
 
 struct pl_color_map_params {
@@ -247,9 +247,16 @@ struct pl_color_map_params {
     float max_boost;
 
     // If true, enables the gamut warning feature. This will visibly highlight
-    // all out-of-gamut colors (by inverting them), if they would have been
+    // all out-of-gamut colors (by coloring them pink), if they would have been
     // clipped as a result of gamut or tone mapping.
     bool gamut_warning;
+
+    // If true, enables colorimetric clipping. This will colorimetrically clip
+    // out-of-gamut colors by desaturating them until they hit the boundary of
+    // the permissible color volume, rather than by hard-clipping. This mode of
+    // clipping preserves luminance between the source and the destination, at
+    // the cost of introducing some color distortion in the opposite direction.
+    bool gamut_clipping;
 };
 
 extern const struct pl_color_map_params pl_color_map_default_params;
@@ -307,6 +314,8 @@ enum pl_dither_method {
     // noise spectrum. Used as a fallback when the above methods are not
     // available.
     PL_DITHER_WHITE_NOISE,
+
+    PL_DITHER_METHOD_COUNT,
 };
 
 struct pl_dither_params {
@@ -347,75 +356,4 @@ void pl_shader_dither(struct pl_shader *sh, int new_depth,
                       struct pl_shader_obj **dither_state,
                       const struct pl_dither_params *params);
 
-struct pl_3dlut_params {
-    // The rendering intent to use when computing the color transformation. A
-    // recommended value is PL_INTENT_RELATIVE_COLORIMETRIC for color-accurate
-    // video reproduction, or PL_INTENT_PERCEPTUAL for profiles containing
-    // meaningful perceptual mapping tables.
-    enum pl_rendering_intent intent;
-
-    // The size of the 3DLUT to generate. If left as NULL, these individually
-    // default to 64, which is the recommended default for all three.
-    size_t size_r, size_g, size_b;
-};
-
-extern const struct pl_3dlut_params pl_3dlut_default_params;
-
-struct pl_3dlut_profile {
-    // The nominal, closest approximation representation of the color profile,
-    // as permitted by `pl_color_space` enums. This will be used as a fallback
-    // in the event that an ICC profile is absent, or that parsing the ICC
-    // profile fails. This is also that will be returned for the corresponding
-    // field in `pl_3dlut_result` when the ICC profile is in use.
-    struct pl_color_space color;
-
-    // The ICC profile itself. (Optional)
-    struct pl_icc_profile profile;
-};
-
-struct pl_3dlut_result {
-    // The source color space. This is the color space that the colors should
-    // actually be in at the point in time that they're ingested by the 3DLUT.
-    // This may differ from the `pl_color_space color` specified in the
-    // `pl_color_profile`. Users should make sure to apply
-    // `pl_shader_color_map` in order to get the colors into this format before
-    // applying `pl_shader_3dlut`.
-    //
-    // Note: `pl_shader_color_map` is a no-op when the source and destination
-    // color spaces are the same, so this can safely be used without disturbing
-    // the colors in the event that an ICC profile is actually in use.
-    struct pl_color_space src_color;
-
-    // The destination color space. This is the color space that the colors
-    // will (nominally) be in at the time they exit the 3DLUT.
-    struct pl_color_space dst_color;
-};
-
-#ifdef PL_HAVE_LCMS
-
-// Updates/generates a 3DLUT. Returns success. If true, `out` will be updated
-// to a struct describing the color space chosen for the input and output of
-// the 3DLUT. (See `pl_color_profile`)
-// If `params` is NULL, it defaults to &pl_3dlut_default_params.
-//
-// Note: This function must always be called before `pl_shader_3dlut`, on the
-// same `pl_shader` object, The only reason it's separate from `pl_shader_3dlut`
-// is to give users a chance to adapt the input colors to the color space
-// chosen by the 3DLUT before applying it.
-bool pl_3dlut_update(struct pl_shader *sh,
-                     const struct pl_3dlut_profile *src,
-                     const struct pl_3dlut_profile *dst,
-                     struct pl_shader_obj **lut3d,
-                     struct pl_3dlut_result *out,
-                     const struct pl_3dlut_params *params);
-
-// Actually applies a 3DLUT as generated by `pl_3dlut_update`. The reason this
-// is separated from `pl_3dlut_update` is so that the user has the chance to
-// correctly map the colors into the specified `src_color` space. This should
-// be called only on the `pl_shader_obj` previously updated by
-// `pl_3dlut_update`, and only when that function returned true.
-void pl_3dlut_apply(struct pl_shader *sh, struct pl_shader_obj **lut3d);
-
-#endif // PL_HAVE_LCMS
-
 #endif // LIBPLACEBO_SHADERS_COLORSPACE_H_
