@@ -39,7 +39,7 @@ static inline enum pl_color_system pl_system_from_av(enum AVColorSpace spc)
     case AVCOL_SPC_RESERVED:            return PL_COLOR_SYSTEM_UNKNOWN;
     case AVCOL_SPC_FCC:                 return PL_COLOR_SYSTEM_UNKNOWN; // missing
     case AVCOL_SPC_BT470BG:             return PL_COLOR_SYSTEM_BT_601;
-    case AVCOL_SPC_SMPTE170M:           return PL_COLOR_SYSTEM_SMPTE_240M;
+    case AVCOL_SPC_SMPTE170M:           return PL_COLOR_SYSTEM_BT_601;
     case AVCOL_SPC_SMPTE240M:           return PL_COLOR_SYSTEM_SMPTE_240M;
     case AVCOL_SPC_YCGCO:               return PL_COLOR_SYSTEM_YCGCO;
     case AVCOL_SPC_BT2020_NCL:          return PL_COLOR_SYSTEM_BT_2020_NC;
@@ -61,7 +61,7 @@ static inline enum AVColorSpace pl_system_to_av(enum pl_color_system sys)
 {
     switch (sys) {
     case PL_COLOR_SYSTEM_UNKNOWN:       return AVCOL_SPC_UNSPECIFIED;
-    case PL_COLOR_SYSTEM_BT_601:        return AVCOL_SPC_BT470BG;
+    case PL_COLOR_SYSTEM_BT_601:        return AVCOL_SPC_SMPTE170M;
     case PL_COLOR_SYSTEM_BT_709:        return AVCOL_SPC_BT709;
     case PL_COLOR_SYSTEM_SMPTE_240M:    return AVCOL_SPC_SMPTE240M;
     case PL_COLOR_SYSTEM_BT_2020_NC:    return AVCOL_SPC_BT2020_NCL;
@@ -246,7 +246,12 @@ static inline int pl_plane_data_from_pixfmt(struct pl_plane_data out_data[4],
                                             enum AVPixelFormat pix_fmt)
 {
     const AVPixFmtDescriptor *desc = av_pix_fmt_desc_get(pix_fmt);
-    assert(desc);
+    int planes = av_pix_fmt_count_planes(pix_fmt);
+    struct pl_plane_data aligned_data[4];
+    struct pl_bit_encoding bits;
+    bool first;
+    if (!desc || planes < 0) // e.g. AV_PIX_FMT_NONE
+        return 0;
 
     if (desc->flags & AV_PIX_FMT_FLAG_BE) {
         // Big endian formats are almost definitely not supported in any
@@ -264,23 +269,28 @@ static inline int pl_plane_data_from_pixfmt(struct pl_plane_data out_data[4],
         return 0;
     }
 
+    if (desc->flags & AV_PIX_FMT_FLAG_BAYER) {
+        // Bayer format don't have valid `desc->offset` values, so we can't
+        // use `pl_plane_data_from_mask` on them.
+        return 0;
+    }
+
     if (desc->nb_components == 0 || desc->nb_components > 4) {
         // Bogus components, possibly fake/virtual/hwaccel format?
         return 0;
     }
 
-    int planes = av_pix_fmt_count_planes(pix_fmt);
     if (planes > 4)
         return 0; // This shouldn't ever happen
 
     // Fill in the details for each plane
     for (int p = 0; p < planes; p++) {
         struct pl_plane_data *data = &out_data[p];
+        uint64_t masks[4] = {0};
         data->type = (desc->flags & AV_PIX_FMT_FLAG_FLOAT)
                         ? PL_FMT_FLOAT
                         : PL_FMT_UNORM;
 
-        uint64_t masks[4] = {0};
         data->pixel_stride = 0;
 
         for (int c = 0; c < desc->nb_components; c++) {
@@ -288,7 +298,7 @@ static inline int pl_plane_data_from_pixfmt(struct pl_plane_data out_data[4],
             if (comp->plane != p)
                 continue;
 
-            masks[c] = (1 << comp->depth) - 1; // e.g. 0xFF for depth=8
+            masks[c] = (1LLU << comp->depth) - 1; // e.g. 0xFF for depth=8
             masks[c] <<= comp->shift;
             masks[c] <<= comp->offset * 8;
 
@@ -307,21 +317,18 @@ static inline int pl_plane_data_from_pixfmt(struct pl_plane_data out_data[4],
         return planes;
 
     // Attempt aligning all of the planes for optimum compatibility
-    struct pl_plane_data data[4];
-    struct pl_bit_encoding bits;
-    bool first = true;
-
+    first = true;
     for (int p = 0; p < planes; p++) {
-        data[p] = out_data[p];
+        aligned_data[p] = out_data[p];
 
         // Planes with only an alpha component should be ignored
-        bool is_alpha = pl_plane_data_num_comps(&data[p]) == 1 &&
-                        data[p].component_map[0] == PL_CHANNEL_A;
-
-        if (is_alpha)
+        if (pl_plane_data_num_comps(&aligned_data[p]) == 1 &&
+            aligned_data[p].component_map[0] == PL_CHANNEL_A)
+        {
             continue;
+        }
 
-        if (!pl_plane_data_align(&data[p], &bits))
+        if (!pl_plane_data_align(&aligned_data[p], &bits))
             goto misaligned;
 
         if (first) {
@@ -335,7 +342,7 @@ static inline int pl_plane_data_from_pixfmt(struct pl_plane_data out_data[4],
 
     // Overwrite the planes by their aligned versions
     for (int p = 0; p < planes; p++)
-        out_data[p] = data[p];
+        out_data[p] = aligned_data[p];
 
     return planes;
 
@@ -363,11 +370,12 @@ static inline bool pl_test_pixfmt(const struct pl_gpu *gpu,
 
 static inline void pl_avframe_set_color(AVFrame *frame, struct pl_color_space space)
 {
+    const AVFrameSideData *sd;
+
     frame->color_primaries = pl_primaries_to_av(space.primaries);
     frame->color_trc = pl_transfer_to_av(space.transfer);
     // No way to map space.light, so just ignore it
 
-    const AVFrameSideData *sd;
     if (space.sig_peak > 1 && space.sig_peak < pl_color_transfer_nominal_peak(space.transfer)) {
         sd = av_frame_get_side_data(frame, AV_FRAME_DATA_CONTENT_LIGHT_LEVEL);
         if (!sd) {
@@ -400,12 +408,12 @@ static inline void pl_avframe_set_repr(AVFrame *frame, struct pl_color_repr repr
 
 static inline void pl_avframe_set_profile(AVFrame *frame, struct pl_icc_profile profile)
 {
+    const AVFrameSideData *sd;
     av_frame_remove_side_data(frame, AV_FRAME_DATA_ICC_PROFILE);
 
     if (!profile.len)
         return;
 
-    const AVFrameSideData *sd;
     sd = av_frame_new_side_data(frame, AV_FRAME_DATA_ICC_PROFILE, profile.len);
     memcpy(sd->data, profile.data, profile.len);
 }
@@ -415,6 +423,7 @@ static inline void pl_frame_from_avframe(struct pl_frame *out,
 {
     const AVPixFmtDescriptor *desc = av_pix_fmt_desc_get(frame->format);
     int planes = av_pix_fmt_count_planes(frame->format);
+    const AVFrameSideData *sd;
     assert(desc);
 
     if (desc->flags & AV_PIX_FMT_FLAG_HWACCEL) {
@@ -483,7 +492,6 @@ static inline void pl_frame_from_avframe(struct pl_frame *out,
         out->repr.sys = pl_color_system_guess_ycbcr(frame->width, frame->height);
     }
 
-    const AVFrameSideData *sd;
     if ((sd = av_frame_get_side_data(frame, AV_FRAME_DATA_ICC_PROFILE))) {
         out->profile = (struct pl_icc_profile) {
             .data = sd->data,
@@ -551,10 +559,8 @@ static inline void pl_frame_from_avframe(struct pl_frame *out,
 
         // Fill in the component mapping array
         for (int c = 0; c < desc->nb_components; c++) {
-            if (desc->comp[c].plane != p)
-                continue;
-
-            plane->component_mapping[plane->components++] = c;
+            if (desc->comp[c].plane == p)
+                plane->component_mapping[plane->components++] = c;
         }
 
         // Clear the superfluous components
@@ -570,33 +576,113 @@ static inline void pl_frame_from_avframe(struct pl_frame *out,
     }
 }
 
-static inline bool pl_upload_avframe(const struct pl_gpu *gpu,
-                                     struct pl_frame *out,
-                                     const struct pl_tex *tex[4],
-                                     const AVFrame *frame)
+static inline bool pl_frame_recreate_from_avframe(const struct pl_gpu *gpu,
+                                                  struct pl_frame *out,
+                                                  const struct pl_tex *tex[4],
+                                                  const AVFrame *frame)
 {
-    pl_frame_from_avframe(out, frame);
-
-    // TODO: support HW-accelerated formats (e.g. wrapping vulkan frames,
-    // importing DRM buffers, and so on)
-
+    const AVPixFmtDescriptor *desc = av_pix_fmt_desc_get(frame->format);
     struct pl_plane_data data[4] = {0};
-    int planes = pl_plane_data_from_pixfmt(data, &out->repr.bits, frame->format);
+    int planes;
+
+    pl_frame_from_avframe(out, frame);
+    planes = pl_plane_data_from_pixfmt(data, &out->repr.bits, frame->format);
     if (!planes)
         return false;
 
-    const AVPixFmtDescriptor *desc = av_pix_fmt_desc_get(frame->format);
     for (int p = 0; p < planes; p++) {
         bool is_chroma = p == 1 || p == 2; // matches lavu logic
         data[p].width = frame->width >> (is_chroma ? desc->log2_chroma_w : 0);
         data[p].height = frame->height >> (is_chroma ? desc->log2_chroma_h : 0);
-        data[p].row_stride = frame->linesize[p];
-        data[p].pixels = frame->data[p];
-        if (!pl_upload_plane(gpu, &out->planes[p], &tex[p], &data[p]))
+
+        if (!pl_recreate_plane(gpu, &out->planes[p], &tex[p], &data[p]))
             return false;
     }
 
     return true;
 }
 
+static void pl_avframe_free(void *priv)
+{
+    AVFrame *frame = priv;
+    av_frame_free(&frame);
+}
+
+static inline bool pl_upload_avframe(const struct pl_gpu *gpu,
+                                     struct pl_frame *out,
+                                     const struct pl_tex *tex[4],
+                                     const AVFrame *frame)
+{
+    const AVPixFmtDescriptor *desc = av_pix_fmt_desc_get(frame->format);
+    struct pl_plane_data data[4] = {0};
+    int planes;
+
+    pl_frame_from_avframe(out, frame);
+
+    // TODO: support HW-accelerated formats (e.g. wrapping vulkan frames,
+    // importing DRM buffers, and so on)
+
+    planes = pl_plane_data_from_pixfmt(data, &out->repr.bits, frame->format);
+    if (!planes)
+        return false;
+
+    for (int p = 0; p < planes; p++) {
+        bool is_chroma = p == 1 || p == 2; // matches lavu logic
+        data[p].width = frame->width >> (is_chroma ? desc->log2_chroma_w : 0);
+        data[p].height = frame->height >> (is_chroma ? desc->log2_chroma_h : 0);
+        data[p].row_stride = frame->linesize[p];
+        data[p].pixels = frame->data[p];
+
+        if (gpu->caps & PL_GPU_CAP_CALLBACKS) {
+            // Use asynchronous upload if possible
+            data[p].callback = pl_avframe_free;
+            data[p].priv = av_frame_clone(frame);
+        }
+
+        if (!pl_upload_plane(gpu, &out->planes[p], &tex[p], &data[p])) {
+            pl_avframe_free(data[p].priv);
+            return false;
+        }
+    }
+
+    return true;
+}
+
+static void pl_done_cb(void *priv)
+{
+    bool *status = priv;
+    *status = true;
+}
+
+static inline bool pl_download_avframe(const struct pl_gpu *gpu,
+                                       const struct pl_frame *frame,
+                                       AVFrame *out_frame)
+{
+    bool done[4] = {0};
+    if (frame->num_planes != av_pix_fmt_count_planes(out_frame->format))
+        return false;
+
+    for (int p = 0; p < frame->num_planes; p++) {
+        size_t texel_size = frame->planes[p].texture->params.format->texel_size;
+        bool ok = pl_tex_download(gpu, &(struct pl_tex_transfer_params) {
+            .tex = frame->planes[p].texture,
+            .stride_w = out_frame->linesize[p] / texel_size,
+            .ptr = out_frame->data[p],
+            // Use synchronous transfer for the last plane
+            .callback = (p+1) < frame->num_planes ? pl_done_cb : NULL,
+            .priv = &done[p],
+        });
+
+        if (!ok)
+            return false;
+    }
+
+    for (int p = 0; p < frame->num_planes - 1; p++) {
+        while (!done[p])
+            pl_tex_poll(gpu, frame->planes[p].texture, UINT64_MAX);
+    }
+
+    return true;
+}
+
 #endif // LIBPLACEBO_LIBAV_H_
